@@ -12,9 +12,10 @@ import numpy as np
 from functools import wraps
 from itertools import combinations, product
 from scipy.spatial import Voronoi
-from scipy.spatial.distance import cdist
-from rpack import pack
+from scipy.spatial.distance import cdist, pdist, squareform
+from scipy.optimize import minimize, NonlinearConstraint
 
+from rpack import pack
 from grandalf.graphs import Vertex, Edge, Graph
 from grandalf.layouts import SugiyamaLayout, DummyVertex
 
@@ -27,6 +28,7 @@ from ._utils import (
     _invert_dict,
     _get_connected_components,
     _convert_polar_to_cartesian_coordinates,
+    _get_angle,
 )
 
 
@@ -273,8 +275,9 @@ def _get_side_by_side_component_bboxes(components, origin, scale, pad_by=0.05):
 def get_fruchterman_reingold_layout(edges,
                                     edge_weights        = None,
                                     k                   = None,
-                                    scale               = None,
-                                    origin              = None,
+                                    origin              = (0, 0),
+                                    scale               = (1, 1),
+                                    pad_by              = 0.05,
                                     initial_temperature = 1.,
                                     total_iterations    = 50,
                                     node_size           = 0,
@@ -296,12 +299,23 @@ def get_fruchterman_reingold_layout(edges,
         Mapping of edges to edge weights.
     k : float or None, default None
         Expected mean edge length. If None, initialized to the sqrt(area / total nodes).
-    origin : tuple or None, default None
+    origin : tuple, default (0, 0)
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
-        If None is given, the origin is placed at (0, 0).
-    scale : tuple or None, default None
+    scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
-        If None is given, the scale is set to (1, 1).
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     total_iterations : int, default 50
         Number of iterations.
     initial_temperature: float, default 1.
@@ -339,27 +353,8 @@ def get_fruchterman_reingold_layout(edges,
     # This wrapper handles the initialization of variables to their defaults (if not explicitely provided),
     # and checks inputs for self-consistency.
 
-    if origin is None:
-        if node_positions:
-            minima = np.min(list(node_positions.values()), axis=0)
-            origin = np.min(np.stack([minima, np.zeros_like(minima)], axis=0), axis=0)
-        else:
-            origin = np.zeros((2))
-    else:
-        # ensure that it is an array
-        origin = np.array(origin)
-
-    if scale is None:
-        if node_positions:
-            delta = np.array(list(node_positions.values())) - origin[np.newaxis, :]
-            maxima = np.max(delta, axis=0)
-            scale = np.max(np.stack([maxima, np.ones_like(maxima)], axis=0), axis=0)
-        else:
-            scale = np.ones((2))
-    else:
-        # ensure that it is an array
-        scale = np.array(scale)
-
+    origin = np.array(origin)
+    scale = np.array(scale)
     assert len(origin) == len(scale), \
         "Arguments `origin` (d={}) and `scale` (d={}) need to have the same number of dimensions!".format(len(origin), len(scale))
     dimensionality = len(origin)
@@ -475,7 +470,7 @@ def get_fruchterman_reingold_layout(edges,
     node_positions_as_array[is_mobile] = mobile_positions
 
     if np.all(is_mobile):
-        node_positions_as_array = _rescale_to_frame(node_positions_as_array, origin, scale)
+        node_positions_as_array = _fit_to_frame(node_positions_as_array, origin, scale, pad_by)
 
     node_positions = dict(zip(unique_nodes, node_positions_as_array))
 
@@ -588,7 +583,7 @@ def get_random_layout(edges, origin=(0,0), scale=(1,1)):
 
 
 @_handle_multiple_components
-def get_sugiyama_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_iterations=3):
+def get_sugiyama_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_size=3, total_iterations=3):
     """'Dot' or Sugiyama node layout.
 
     Uses the Sugiyama algorithm [Sugiyama1981]_ to compute node positions.
@@ -598,12 +593,23 @@ def get_sugiyama_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_ite
     ----------
     edges : list
         The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
-    origin : tuple or None, default None
+    origin : tuple, default (0, 0)
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
-        If None is given, the origin is placed at (0, 0).
-    scale : tuple or None, default None
+    scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
-        If None is given, the scale is set to (1, 1).
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     total_iterations : int, default 50
         Number of iterations.
 
@@ -641,7 +647,7 @@ def get_sugiyama_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_ite
     # rescale to canvas
     # TODO: by rescaling, we effectively ignore the node_size argument
     nodes, positions = zip(*node_positions.items())
-    positions = _rescale_to_frame(np.array(positions), np.array(origin), np.array(scale))
+    positions = _rescale_to_frame(np.array(positions), np.array(origin) + pad_by * np.array(scale), (1 - pad_by) * np.array(scale))
 
     # place roots on top, leaves on bottom
     positions[:, 1] -= origin[1] + scale[1]
@@ -680,7 +686,7 @@ class vertex_view(object):
 
 
 @_handle_multiple_components
-def get_radial_tree_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_iterations=3):
+def get_radial_tree_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_size=3, total_iterations=3):
     """Radial tree layout.
 
     Uses the Sugiyama algorithm [Sugiyama1981]_ to compute node positions.
@@ -690,12 +696,23 @@ def get_radial_tree_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_
     ----------
     edges : list
         The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
-    origin : tuple or None, default None
+    origin : tuple, default (0, 0)
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
-        If None is given, the origin is placed at (0, 0).
-    scale : tuple or None, default None
+    scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
-        If None is given, the scale is set to (1, 1).
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     total_iterations : int, default 50
         Number of iterations.
 
@@ -711,12 +728,12 @@ def get_radial_tree_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_
 
     """
 
-    sugiyama_positions = get_sugiyama_layout(
-        edges,
-        origin=(0,0),
-        scale=(1,1),
-        node_size=node_size,
-        total_iterations=total_iterations
+    sugiyama_positions = get_sugiyama_layout(edges,
+                                             origin           = (0, 0),
+                                             scale            = (1, 1),
+                                             pad_by           = 0,
+                                             node_size        = node_size,
+                                             total_iterations = total_iterations
     )
 
     # determine the size of the largest layer
@@ -734,7 +751,7 @@ def get_radial_tree_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_
 
     max_angle = 2 * np.pi * max_nodes_per_layer / (max_nodes_per_layer + 1)
     max_radius = np.min(scale) / 2
-    max_radius *= 0.9 # shrink to make room for node artists, labels, and annotations
+    max_radius *= (1 - pad_by) # shrink to make room for node artists, labels, and annotations
     offset = np.array([origin[0] + 0.5 * scale[0], origin[1] + 0.5 * scale[1]])
 
     node_positions = dict()
@@ -747,7 +764,7 @@ def get_radial_tree_layout(edges, origin=(0,0), scale=(1,1), node_size=3, total_
 
 
 @_handle_multiple_components
-def get_circular_layout(edges, origin=(0,0), scale=(1,1), node_order=None, reduce_edge_crossings=True):
+def get_circular_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_order=None, reduce_edge_crossings=True):
     """Circular node layout.
 
     By default, this implementation uses a heuristic to arrange the nodes such that the edge crossings are minimised.
@@ -756,10 +773,23 @@ def get_circular_layout(edges, origin=(0,0), scale=(1,1), node_order=None, reduc
     ----------
     edges : list
         The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
-    origin : tuple
+    origin : tuple, default (0, 0)
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
-    scale : tuple
+    scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     node_order : list or None, default None
         The (initial) ordering of nodes (counter-clockwise) before layout optimisation.
         Set :code:`reduce_edge_crossings` to :code:`False` to skip optimisation and retain the given node order in the returned layout.
@@ -779,7 +809,7 @@ def get_circular_layout(edges, origin=(0,0), scale=(1,1), node_order=None, reduc
     nodes = _get_unique_nodes(edges)
     center = np.array(origin) + 0.5 * np.array(scale)
     radius = np.min(scale) / 2
-    radius *= 0.9 # fudge factor to make space for self-loops, annotations, etc
+    radius *= (1 - pad_by) # fudge factor to make space for self-loops, annotations, etc
     positions = _get_n_points_on_a_circle(center, radius, len(nodes), start_angle=0)
 
     if node_order:
@@ -1040,7 +1070,7 @@ def _get_centroid(polygon):
 
 
 @_handle_multiple_components
-def get_linear_layout(edges, origin=(0,0), scale=(1,1), node_order=None, reduce_edge_crossings=True):
+def get_linear_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_order=None, reduce_edge_crossings=True):
     """Linear node layout.
 
     If :code:`reduce_edge_crossings` is set to :code:`True`, the algorithm
@@ -1054,6 +1084,19 @@ def get_linear_layout(edges, origin=(0,0), scale=(1,1), node_order=None, reduce_
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
     scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     node_order : list or None, default None
         The (initial) ordering of nodes (counter-clockwise) before layout optimisation.
         Set :code:`reduce_edge_crossings` to :code:`False` to skip optimisation and retain the given node order in the returned layout.
@@ -1074,7 +1117,7 @@ def get_linear_layout(edges, origin=(0,0), scale=(1,1), node_order=None, reduce_
 
     nodes = _get_unique_nodes(edges)
     total_nodes = len(nodes)
-    x = np.linspace(origin[0], origin[0] + scale[0], total_nodes+2)[1:-1]
+    x = np.linspace(origin[0] + pad_by * scale[0], origin[0] + (1 - pad_by) * scale[0], total_nodes)
     y = np.full(total_nodes, origin[1] + 0.5 * scale[1])
     positions = np.c_[x, y]
 
@@ -1110,7 +1153,7 @@ def _minimize_total_edge_length(nodes, edges):
     return output
 
 
-def get_bipartite_layout(edges, nodes=None, subsets=None, origin=(0, 0), scale=(1, 1), reduce_edge_crossings=True):
+def get_bipartite_layout(edges, nodes=None, subsets=None, origin=(0, 0), scale=(1, 1), pad_by=0.05, reduce_edge_crossings=True):
     """Bipartite node layout.
 
     By default, this implementation uses a heuristic to arrange the nodes such that the edge crossings are reduced.
@@ -1126,6 +1169,19 @@ def get_bipartite_layout(edges, nodes=None, subsets=None, origin=(0, 0), scale=(
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
     scale : tuple
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     reduce_edge_crossings : bool, default True
         If True, attempts to reduce edge crossings via the algorithm outlined in [Eades1994]_.
 
@@ -1161,6 +1217,10 @@ def get_bipartite_layout(edges, nodes=None, subsets=None, origin=(0, 0), scale=(
         if not _is_complete_bipartite(edges, left, right):
             left, right = _reduce_crossings_bipartite(adjacency_list, left, right)
 
+    # shrink frame to apply padding
+    origin = np.array(origin) + pad_by * np.array(scale)
+    scale = np.array(scale) * (1 - 2 * pad_by)
+
     if len(left) > len(right):
         spacing = scale[1] / (len(left) - 1)
     else:
@@ -1244,7 +1304,7 @@ def _reduce_crossings_bipartite(adjacency_list, left, right):
     return left, sorted(right_ranks, key=right_ranks.get)
 
 
-def get_multipartite_layout(edges, layers, layer_positions=None, origin=(0, 0), scale=(1, 1), reduce_edge_crossings=True, uniform_node_spacing=True):
+def get_multipartite_layout(edges, layers, layer_positions=None, origin=(0, 0), scale=(1, 1), pad_by=0.05, reduce_edge_crossings=True, uniform_node_spacing=True):
     """Layered node layout for a multipartite graph.
 
     By default, this implementation uses a heuristic to arrange the nodes such that the edge crossings are reduced.
@@ -1262,6 +1322,19 @@ def get_multipartite_layout(edges, layers, layer_positions=None, origin=(0, 0),
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
     scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     reduce_edge_crossings : bool, default True
         If True, attempts to reduce edge crossings via the algorithm outlined in [Eades1994]_.
     uniform_node_spacing : bool, default True
@@ -1279,15 +1352,19 @@ def get_multipartite_layout(edges, layers, layer_positions=None, origin=(0, 0),
 
     """
 
+    # shrink frame to apply padding
+    origin = np.array(origin) + pad_by * np.array(scale)
+    scale = np.array(scale) * (1 - 2 * pad_by)
+
     # set the space between nodes
     if uniform_node_spacing:
         try:
-            node_spacings = scale[1] / (np.max([len(layer) for layer in layers]) - 1) * np.ones_like(layers, dtype=np.float)
+            node_spacings = scale[1] / (np.max([len(layer) for layer in layers]) - 1) * np.ones_like(layers, dtype=float)
         except ZeroDivisionError:
             # The graph has at most a single edge between each pair of layers.
-            node_spacings = np.ones_like(layers, dtype=np.float)
+            node_spacings = np.ones_like(layers, dtype=float)
     else:
-        node_spacings = np.ones_like(layers, dtype=np.float)
+        node_spacings = np.ones_like(layers, dtype=float)
         for ii, layer in enumerate(layers):
             if len(layer) > 1:
                 node_spacings[ii] = 1./(len(layer) - 1)
@@ -1339,7 +1416,7 @@ def _get_node_positions_within_layer(node_order, node_spacing, layer_position, o
     return node_positions
 
 
-def get_shell_layout(edges, shells, radii=None, origin=(0, 0), scale=(1, 1), reduce_edge_crossings=True):
+def get_shell_layout(edges, shells, radii=None, origin=(0, 0), scale=(1, 1), pad_by=0.05, reduce_edge_crossings=True):
     """Shell layout.
 
     This is a wrapper around `get_multipartite_layout` that arranges nodes in shells around a center instead of in layers.
@@ -1357,6 +1434,19 @@ def get_shell_layout(edges, shells, radii=None, origin=(0, 0), scale=(1, 1), red
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
     scale : tuple
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
+
     reduce_edge_crossings : bool, default True
         If True, attempts to reduce edge crossings via the algorithm outlined in [Eades1994]_.
 
@@ -1365,20 +1455,26 @@ def get_shell_layout(edges, shells, radii=None, origin=(0, 0), scale=(1, 1), red
     node_positions : dict
         Dictionary mapping each node ID to (float x, float y) tuple, the node position.
 
+    References
+    ----------
+    .. [Eades1994] Eades & Wormald (1994) Edge crossings in drawings of bipartite graphs.
+
     """
+
     if radii is None:
         if len(shells[0]) == 1:
             # Innermost shell consists of a single node, and hence should have no size.
-            radii = np.linspace(0, 0.9 * np.min(scale) / 2, len(shells))
+            radii = np.linspace(0, (1 - pad_by) * np.min(scale) / 2, len(shells))
         else:
             # Innermost shell consists of multiple nodes and should hence have a non-zero radius.
-            radii = np.linspace(0, 0.9 * np.min(scale) / 2, len(shells) + 1)[1:]
+            radii = np.linspace(0, (1 - pad_by) * np.min(scale) / 2, len(shells) + 1)[1:]
 
     relative_radii = np.array(radii) / np.max(radii)
     multipartite_positions = get_multipartite_layout(
         edges, shells, layer_positions=relative_radii,
         reduce_edge_crossings=reduce_edge_crossings,
-        uniform_node_spacing=False
+        uniform_node_spacing=False,
+        pad_by=0,
     )
 
     node_to_shell = {node : shell for shell in shells for node in shell}
@@ -1398,7 +1494,7 @@ def get_shell_layout(edges, shells, radii=None, origin=(0, 0), scale=(1, 1), red
 
 
 @_handle_multiple_components
-def get_community_layout(edges, node_to_community, origin=(0,0), scale=(1,1)):
+def get_community_layout(edges, node_to_community, origin=(0, 0), scale=(1, 1), pad_by=0.05):
     """Community node layout for modular graphs.
 
     This implements the following steps:
@@ -1422,6 +1518,18 @@ def get_community_layout(edges, node_to_community, origin=(0,0), scale=(1,1)):
         The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
     scale : tuple, default (1, 1)
         The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+
+        :code:`xmin = origin[0] + pad_by * scale[0]`
+
+        :code:`xmax = origin[0] + scale[0] - pad_by * scale[0]`
+
+        :code:`ymin = origin[1] + pad_by * scale[1]`
+
+        :code:`ymax = origin[1] + scale[1] - pad_by * scale[1]`
 
     Returns
     -------
@@ -1435,14 +1543,14 @@ def get_community_layout(edges, node_to_community, origin=(0,0), scale=(1,1)):
     communities = set([node_to_community[node] for node in nodes])
     if len(communities) < 2:
         warnings.warn("Graph contains a single community. Unable to compute a community layout. Computing spring layout instead.")
-        return get_fruchterman_reingold_layout(edges, origin=origin, scale=scale)
+        return get_fruchterman_reingold_layout(edges, origin=origin, scale=scale, pad_by=pad_by)
 
     # assert that node_to_community is non-redundant,
     # i.e. only contains nodes that are also present in edges
     node_to_community = {node : node_to_community[node] for node in nodes}
 
     community_size = _get_community_sizes(node_to_community, scale)
-    community_centroids = _get_community_positions(edges, node_to_community, community_size, origin, scale)
+    community_centroids = _get_community_positions(edges, node_to_community, community_size, origin, scale, pad_by)
     relative_node_positions = _get_within_community_positions(edges, node_to_community)
     node_positions = _combine_positions(node_to_community, community_centroids, community_size, relative_node_positions)
     node_positions = _rotate_communities(edges, node_to_community, community_centroids, node_positions)
@@ -1454,13 +1562,13 @@ def _get_community_sizes(node_to_community, scale):
     """Compute the area of the canvas reserved for each community."""
     total_nodes = len(node_to_community)
     max_radius = np.linalg.norm(scale) / 2
-    scalar = max_radius / total_nodes
+    scalar = max_radius / total_nodes # this is the worst case scenario, where all comunities are lined up like beads on a string; may warrant revisiting
     community_to_nodes = _invert_dict(node_to_community)
     community_size = {community : len(nodes) * scalar for community, nodes in community_to_nodes.items()}
     return community_size
 
 
-def _get_community_positions(edges, node_to_community, community_size, origin, scale):
+def _get_community_positions(edges, node_to_community, community_size, origin, scale, pad_by):
     """Compute a centroid position for each community."""
     # create a weighted graph, in which each node corresponds to a community,
     # and each edge weight to the number of edges between communities
@@ -1469,7 +1577,7 @@ def _get_community_positions(edges, node_to_community, community_size, origin, s
     # find layout for communities
     return get_fruchterman_reingold_layout(
         list(between_community_edges.keys()), edge_weight=between_community_edges,
-        node_size=community_size, origin=origin, scale=scale,
+        node_size=community_size, origin=origin, scale=scale, pad_by=pad_by,
     )
 
 
@@ -1521,7 +1629,7 @@ def _combine_positions(node_to_community, community_centroids, community_size, r
     return node_positions
 
 
-def _rotate_communities(edges, node_to_community, community_centroids, node_positions, step_size=-0.01, max_iterations=100):
+def _rotate_communities(edges, node_to_community, community_centroids, node_positions, step_size=0.1, max_iterations=200):
 
     between_community_edges = [(source, target) for (source, target) in edges \
                                if node_to_community[source] != node_to_community[target]]
@@ -1545,12 +1653,9 @@ def _rotate_communities(edges, node_to_community, community_centroids, node_posi
             F = delta * np.linalg.norm(delta)
             community_torque[community] += np.cross(r, F)
 
-        # TODO: compare new torque values to previous; abort if change is small
-
         # update node positions
-        step_size = -0.1
         for node, community in node_to_community.items():
-            node_positions[node] = _rotate(step_size * community_torque[community],
+            node_positions[node] = _rotate(step_size * -community_torque[community],
                                            node_positions[node],
                                            community_centroids[community])
 
@@ -1564,3 +1669,184 @@ def _rotate(angle, points, origin=(0, 0)):
     origin = np.atleast_2d(origin)
     points = np.atleast_2d(points)
     return np.squeeze((R @ (points.T-origin.T) + origin.T).T)
+
+
+@_handle_multiple_components
+def get_geometric_layout(edges, edge_length, node_size=0., tol=1e-3, origin=(0, 0), scale=(1, 1), pad_by=0.05):
+    """Node layout for defined edge lengths but unknown node positions.
+
+    Node positions are determined through non-linear optimisation: the
+    total distance between nodes is maximised subject to the constraint
+    imposed by the edge lengths, which are used as upper bounds.
+    If provided, node sizes are used to set lower bounds to minimise collisions.
+
+    ..note:: This implementation is slow.
+
+    Parameters
+    ----------
+    edges : list
+        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
+    edge_lengths : dict
+        Mapping of edges to their lengths.
+    node_size : scalar or dict, default 0.
+        Size (radius) of nodes.
+        Providing the correct node size minimises the overlap of nodes in the graph,
+        which can otherwise occur if there are many nodes, or if the nodes differ considerably in size.
+    tolerance : float, default 1e-3
+        The tolerance of the cost function. Small values increase the accuracy, large values improve the computation time.
+    origin : tuple, default (0, 0)
+        The (float x, float y) coordinates corresponding to the lower left hand corner of the bounding box specifying the extent of the canvas.
+    scale : tuple, default (1, 1)
+        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
+    pad_by : float, default 0.05
+        Padding around node positions to reduce clipping of the node artists with the frame,
+        and to create space for routing curved edges including self-loops around nodes.
+        This results in the following bounding box:
+        xmin, xmax = origin[0] + pad_by * scale[0], origin[0] + scale[0] - pad_by * scale[0]
+        ymin, ymax = origin[1] + pad_by * scale[1], origin[1] + scale[1] - pad_by * scale[1]
+
+    Returns
+    -------
+    node_positions : dict
+        Dictionary mapping each node ID to (float x, float y) tuple, the node position.
+
+    """
+
+    # TODO: assert triangle inequality is not violated.
+    # HOLD: probably not necessary, as minimisation can still proceed when triangle inequality is violated.
+
+    # assert that the edges fit within the canvas dimensions
+    width, height = scale
+    max_length = np.sqrt(width**2 + height**2)
+    too_long = dict()
+    for edge, length in edge_length.items():
+        if length > max_length:
+            too_long[edge] = length
+    if too_long:
+        msg = f"The following edges exceed the dimensions of the canvas (`scale={scale}`):"
+        for edge, length in too_long.items():
+            msg += f"\n\t{edge} : {length}"
+        msg += "\nEither increase the `scale` parameter, or decrease the edge lengths."
+        raise ValueError(msg)
+
+    # ensure that graph is bi-directional
+    edges = edges + [(target, source) for (source, target) in edges] # forces copy
+    edges = list(set(edges))
+
+    # upper bound: pairwise distance matrix with unknown distances set to the maximum possible distance given the canvas dimensions
+
+    lengths = []
+    for (source, target) in edges:
+        if (source, target) in edge_length:
+            lengths.append(edge_length[(source, target)])
+        else:
+            lengths.append(edge_length[(target, source)])
+
+    sources, targets = zip(*edges)
+    nodes = sources + targets
+    unique_nodes = set(nodes)
+    indices = range(len(unique_nodes))
+    node_to_idx = dict(zip(unique_nodes, indices))
+    source_indices = [node_to_idx[source] for source in sources]
+    target_indices = [node_to_idx[target] for target in targets]
+
+    total_nodes = len(unique_nodes)
+    distance_matrix = np.full((total_nodes, total_nodes), max_length)
+    distance_matrix[source_indices, target_indices] = lengths
+    distance_matrix[np.diag_indices(total_nodes)] = 0
+    upper_bounds = squareform(distance_matrix)
+
+    # lower bound: sum of node sizes
+    if isinstance(node_size, (int, float)):
+        sizes = node_size * np.ones((total_nodes))
+    elif isinstance(node_size, dict):
+        sizes = np.array([node_size[node] if node in node_size else 0. for node in unique_nodes])
+
+    sum_of_node_sizes = sizes[np.newaxis, :] + sizes[:, np.newaxis]
+    sum_of_node_sizes -= np.diag(np.diag(sum_of_node_sizes)) # squareform requires zeros on diagonal
+    lower_bounds = squareform(sum_of_node_sizes)
+    invalid = lower_bounds > upper_bounds
+    lower_bounds[invalid] = upper_bounds[invalid] - 1e-8
+
+    # For an extended discussion of this cost function and alternatives see:
+    # https://stackoverflow.com/q/75137677/2912349
+    def cost_function(positions):
+        return 1 / np.sum(np.log(pdist(positions.reshape((-1, 2))) + 1))
+
+    def constraint_function(positions):
+        positions = np.reshape(positions, (-1, 2))
+        return pdist(positions)
+
+    initial_positions = _initialise_geometric_node_layout(edges, edge_length)
+    nonlinear_constraint = NonlinearConstraint(constraint_function, lb=lower_bounds, ub=upper_bounds, jac='2-point')
+    result = minimize(
+        cost_function,
+        initial_positions.flatten(),
+        method='SLSQP',
+        jac='2-point',
+        constraints=[nonlinear_constraint],
+        options=dict(ftol=tol),
+    )
+
+    if not result.success:
+        print("Warning: could not compute valid node positions for the given edge lengths.")
+        print(f"scipy.optimize.minimize: {result.message}.")
+
+    node_positions_as_array = result.x.reshape((-1, 2))
+    node_positions_as_array = _fit_to_frame(node_positions_as_array, np.array(origin), np.array(scale), pad_by)
+    node_positions = dict(zip(unique_nodes, node_positions_as_array))
+    return node_positions
+
+
+def _initialise_geometric_node_layout(edges, edge_length):
+    """Initialises the node positions using the FR algorithm with weights.
+    Shorter edges are given a larger weight such that the nodes experience a strong attractive force."""
+
+    edge_weight = dict()
+    for edge, length in edge_length.items():
+        edge_weight[edge] = 1 / length
+    node_positions = get_fruchterman_reingold_layout(edges, edge_weight=edge_weight)
+    return np.array(list(node_positions.values()))
+
+
+def _fit_to_frame(positions, origin, scale, pad_by):
+    """Rotate, rescale and shift a set of positions such that they fit
+    inside a frame while preserving the relative distances between
+    them."""
+
+    # find major axis
+    delta = positions[np.newaxis, :] - positions[:, np.newaxis]
+    distances = np.sum(delta**2, axis=-1)
+    ii, jj = np.where(np.triu(distances)==np.max(distances))
+
+    # use the first if there are several solutions
+    ii = ii[0]
+    jj = jj[0]
+
+    # pivot around half-way point
+    pivot = positions[ii] + 0.5 * delta[ii, jj]
+    angle = _get_angle(*delta[ii, jj])
+
+    if scale[0] < scale[1]: # portrait
+        rotated_positions = _rotate((np.pi/2 - angle) % np.pi, positions, pivot)
+    else: # landscape
+        rotated_positions = _rotate(-angle % np.pi, positions, pivot)
+
+    # shift to (0, 0)
+    shifted_positions = rotated_positions - np.min(rotated_positions, axis=0)[np.newaxis, :]
+
+    # rescale & center
+    dx, dy = np.ptp(rotated_positions, axis=0)
+    if dx/scale[0] < dy/scale[1]:
+        rescaled_positions = shifted_positions * (1 - 2 * pad_by) * scale[1] / dy
+        rescaled_positions[:, 0] += (scale[0] - np.ptp(rescaled_positions[:, 0])) / 2
+        rescaled_positions[:, 1] += pad_by * scale[1]
+    else:
+        rescaled_positions = shifted_positions * (1 - 2 * pad_by) * scale[0] / dx
+        rescaled_positions[:, 0] += pad_by * scale[0]
+        rescaled_positions[:, 1] += (scale[1] - np.ptp(rescaled_positions[:, 1])) / 2
+
+    # shift to origin
+    reshifted_positions = rescaled_positions + np.array(origin)[np.newaxis, :]
+
+    return reshifted_positions
